import numpy as np
from sitelle.region import centered_square_region
from sitelle.plot import *
from photutils import CircularAperture, CircularAnnulus, EllipticalAperture, DAOStarFinder, IRAFStarFinder, find_peaks, detect_sources
from scipy.interpolate import UnivariateSpline
from astropy.stats import sigma_clipped_stats
from orb.astrometry import Astrometry
from astropy.convolution import Gaussian2DKernel
from astropy.table import Table
from orb.utils import vector

def filter_sources(sources, annulus):
    x,y = sources['ycentroid'], sources['xcentroid']
    return sources[annulus[np.round(x).astype(int), np.round(y).astype(int)].astype(bool)]

def filter_frame(frame, annulus, val=0):
    _frame = np.copy(frame)
    _frame[np.where(annulus == 0)] = val
    return _frame

def extract_max_frame(x,y, spectral_cube, id_max_detection_frame):
    """
    For a given position, extracts the frames around the maxium of detection in the cube and returns the sum
    :param x: abscissa of the source
    :param y: ordinate of the source
    :param spectral_cube: SpectralCube instance where we are looking at the source
    :param id_max_detection_frame: the id of the max frame, or the detection_pos_frame
    :return: a cutout of the sum of the frames around the max of detection
    """
    if type(id_max_detection_frame) == int:
        iframe = id_max_detection_frame
    elif type(id_max_detection_frame) == np.ndarray:
        iframe = int(id_max_detection_frame[x,y])
    else:
        raise TypeError('Non valid type for id_max_detection_frame : %s'%type(id_max_detection_frame))
    data = spectral_cube.get_data(x-10, x+11, y-10,y+11, iframe-2,iframe+3)
    return np.sum(data, axis=2)

def extract_point_source(xy, cube, small_bin=3, big_bin = 30):
    """
    Basic way to extract a point source spectra with the local background subtracted.
    For a given position xy, we sum the spectra extracted in a squared region of size small_bin**2 centered on xy,
    and subtract from it the median spectra from a squared region of size big_bin**2 centered on xy excluding the central area
    :param xy: tuple of the position
    :param cube: the SpectralCube in which we extract the spectra
    :param small_bin: (Dafault 3) the binsize of the region of extraction of the source
    :param big_bin: (Default 30) the binsize of the region of extraction of the background
    """
    big_box = centered_square_region(*xy, b=big_bin)
    small_box = centered_square_region(*xy, b=small_bin)
    mask = np.zeros((cube.dimx, cube.dimy))
    mask[big_box]=1
    mask[small_box]=0
    _, bkg_spec = cube.extract_integrated_spectrum(np.nonzero(mask), median=True, mean_flux=True, silent=True)
    a,s, n = cube.extract_integrated_spectrum(small_box, silent=True, return_spec_nb = True)
    return a, s-n*bkg_spec

def check_source(x,y, spectral_cube, detection_pos_frame=None, smooth_factor = None):
    """
    Helper function to quickly look at a source
    We extract the source at positon (x,y) with extract_point_source and plot the resulting spectra.
    We also plot a map around the source, to check if we actually detect something
    :param x: abscissa of the source
    :param y: ordinate of the source
    :param spectral_cube: SpectralCube instance where we look at the source
    :param detection_pos_frame: (Optional) If passed, we are going to extract the max frame to plot the detection on top if it. If None, we use the deep_frame
    :param smooth_factor: (Optional) Factor used to smooth the spectrum
    """
    a,s = extract_point_source((x,y), spectral_cube)
    if smooth_factor is not None:
        s = vector.smooth(s, smooth_factor)
    f = plot_spectra(a,s)
    if detection_pos_frame is not None:
        f,ax = plot_map(extract_max_frame(x,y, spectral_cube, detection_pos_frame))
        ax.scatter(10,10, marker='+', color='red')
        wl = 1e8/spectral_cube.params.base_axis[int(detection_pos_frame[x,y])]
        ax.set_title('Frame at %.1f Angstroms'%wl)

    else:
        f,ax = plot_map(spectral_cube.get_deep_frame()[x-10:x+11, y-10:y+11])
        ax.scatter(10,10, marker='+', color='red')

def measure_coherence(source, detection_pos_frame):
    """
    Coherence is a measure of the credibility of a source as an emission line source.
    It checks if the hot pixels of a source are coming from ~the same frames of the cube
    :param source: pandas Dataframe with 'xcentroid', 'ycentroid' columns
    :param detection_pos_frame: map of the id of the max pixels
    """
    y,x = np.round(source[['xcentroid', 'ycentroid']]).astype(int)
    return 1/np.nanstd(detection_pos_frame[x-1:x+2, y-1:y+2])

def measure_source_fwhm(detection, data, rmax=10):
    x,y = np.array(detection[['xcentroid', 'ycentroid']])
    photo_flux = np.zeros(rmax)
    for r in range(rmax):
        if r == 0:
            aper = CircularAperture((x,y), 1.)
        else:
            aper = CircularAnnulus((x,y), r, r+1)
        photo_flux[r] = aper.do_photometry(data)[0]/aper.area()


    def get_fwhm(flux):
        #We assume max is on 0. If not, source is probably contaminated
        flux = flux - flux.min()
        spline = UnivariateSpline(np.arange(rmax), flux-flux[0]/2., s=0)
        if spline.roots().shape != (1,):
            #print spline.roots()
            return np.nan
        return spline.roots()[0]
    return (get_fwhm(photo_flux))


def get_sources(detection_frame, mask=False, sigma = 5.0, mode='DAO', fwhm = 2.5, threshold = None):
    if mask is False:
        mask = np.ones_like(detection_frame)
    mean, median, std = sigma_clipped_stats(detection_frame, sigma=3.0, iters=5,
                                        mask=~mask.astype(bool) )#On masque la region hors de l'anneau
    #On detecte sur toute la frame, mais on garde que ce qui est effectivement dans l'anneau
    if threshold is None:
        threshold = median+sigma*std

    if mode == 'DAO':
        daofind = DAOStarFinder(fwhm=fwhm, threshold=threshold)
        sources = daofind(detection_frame)
    elif mode == 'IRAF':
        irafind = IRAFStarFinder(threshold=threshold, fwhm=fwhm)
        sources = irafind(detection_frame)
    elif mode == 'PEAK':
        sources = find_peaks(detection_frame, threshold=threshold )
        sources.rename_column('x_peak', 'xcentroid')
        sources.rename_column('y_peak', 'ycentroid')
    elif mode == 'ORB':
        astro = Astrometry(detection_frame, instrument='sitelle')
        path, fwhm_arc = astro.detect_stars(min_star_number=5000, r_max_coeff=1. )
        star_list = astro.load_star_list(path)
        sources = Table([star_list[:,0], star_list[:,1]], names=('ycentroid', 'xcentroid'))
    elif mode == 'SEGM':

        s = fwhm / (2.0 * np.sqrt(2.0 * np.log(2.0)))
        kernel = Gaussian2DKernel(s, x_size=3, y_size=3)
        kernel.normalize()

        segm = detect_sources(detection_frame, threshold, npixels=7,
                              filter_kernel=kernel)
        def get_xy(bbox):
            xslice, yslice = bbox
            x = xslice.start + (xslice.stop - xslice.start -1)/2.
            y = yslice.start + (yslice.stop - yslice.start -1)/2.
            return x,y
        def get_positions(segm):
            pos = np.zeros((len(segm.slices), 2))
            for i, box in enumerate(segm.slices):
                pos[i] = np.array(get_xy(box))
            X = pos[:,0]
            Y = pos[:,1]
            return Table([Y, X], names=('xcentroid', 'ycentroid'))

        sources = get_positions(segm)
    sources = filter_sources(sources, mask) # On filtre
    df = sources.to_pandas()
    if 'id' in df:
        df.pop('id')
    return df



from astropy.io import fits
from sitelle.utils import *
import numpy as np
from scipy.interpolate import UnivariateSpline
from orb.utils import io
import subprocess
import os
import copy
import sys
from path import Path
import socket


def parameter_map(table, param, binMap):
    """
    Return a 2D map of the given parameter
    :param table: the BINTable in which the parameters are found
    :param param: The parameter we want to plot. See table.columns to see the list
    :param binMap: the 2D map of the binning scheme
    """
    def f(binNumber):
        return table[param][binNumber]
    return f(binMap)

def read(filename):
    """
    Reads results from NBurst
    :param filename: the file to read
    :return: the bin_table, the fit_table
    """
    hdu = fits.open(filename)
    bin_table = hdu[1].data
    fit_table = hdu[2].data
    hdu.close()
    return bin_table, fit_table

def extract_spectrum(fit_table, binNumber):
    """
    Get the fitted spectrum from a fit8table at a given binNumber
    :param fit_table: the fit_table containg the results
    :param binNumber: the binNumber
    :return: the wavelength axis (in Angstroms) and fitted spectrum
    """
    axis = fit_table['WAVE'][binNumber,:]
    fit = fit_table['FIT'][binNumber,:]
    return axis,fit

def sew_spectra(left, right):
    """
    Sew 2 spectra fitter together.
    In between, we put a mean value for data & fwhm, and 10x the worst error for the error
    :prama left: left spectra fitter
    :param right: right spectra fitter
    :return: a NburstFitter instance containing the 2 spectra
    """
    if left.spectra.shape[:-1] != right.spectra.shape[:-1]:
        raise ValueError('Left and right spectra should have the same spatial dimension. Got %s and %s'%(left.spectra.shape[:-1],right.spectra.shape[:-1]))

    step = max(left.header['CDELT1'], right.header['CDELT1'])#We choose the largest step (lower resolution)
    axis = np.arange(left.axis.min(), right.axis.max(), step=step)
    spectra = np.zeros( left.spectra.shape[:-1] + axis.shape)
    error = np.zeros( left.spectra.shape[:-1] + axis.shape )
    fwhm = np.zeros( left.spectra.shape[:-1] + axis.shape)

    left_max_id = np.searchsorted(axis, left.axis.max(), side='right')
    right_min_id = np.searchsorted(axis, right.axis.min())

    def interpolator(spectrum, old_axis, new_axis):
        f = UnivariateSpline(old_axis, spectrum, s=0)
        return f(new_axis)

    spectra[..., :left_max_id] = np.apply_along_axis(interpolator, -1, left.spectra, left.axis, axis[:left_max_id])
    error[..., :left_max_id] = np.apply_along_axis(interpolator, -1, left.error, left.axis, axis[:left_max_id])
    fwhm[..., :left_max_id] = np.apply_along_axis(interpolator, -1, left.fwhm, left.axis, axis[:left_max_id])

    spectra[..., right_min_id:] = np.apply_along_axis(interpolator, -1, right.spectra, right.axis, axis[right_min_id:])
    error[..., right_min_id:] = np.apply_along_axis(interpolator, -1, right.error, right.axis, axis[right_min_id:])
    fwhm[..., right_min_id:] = np.apply_along_axis(interpolator, -1, right.fwhm, right.axis, axis[right_min_id:])

    # What do we put in between ?
    # Mean for fwhm & data cube
    # 10 x worst error for error
    def spec_func(vect, left_id, right_id):
        left = np.nanmean(vect[:left_id])
        right = np.nanmean(vect[right_id:])
        return np.linspace(left, right, right_id - left_id)
    def fwhm_func(vect, left_id, right_id):
        return np.linspace(vect[left_id-1], vect[right_id], right_id - left_id)
    def err_func(vect, left_id, right_id):
        worst = 10*max(vect[left_id-1], vect[right_id])
        return np.ones(right_id - left_id)*worst

    spectra[..., left_max_id:right_min_id] = np.apply_along_axis(spec_func, -1, spectra, left_max_id, right_min_id)
    error[..., left_max_id:right_min_id] = np.apply_along_axis(err_func, -1, error, left_max_id, right_min_id)
    fwhm[..., left_max_id:right_min_id] = np.apply_along_axis(fwhm_func, -1, fwhm, left_max_id, right_min_id)

    return NburstFitter(axis, spectra, error, fwhm, left.header)

class NburstFitter():
    """
    This class is a helper to fit spectra with Nburst directly from python.
    It stores data as expected by Nburst, configure a script with user-defined parameters, runs it and reads the results.

    """

    nburst_working_dir = Path('/Users/blaunet/Documents/M31/nburst/')
    idl_binary_path = "/usr/local/idl/idl/bin:/usr/local/idl/idl/bin/bin.darwin.x86_64:"
    idl_startup_script = "~/.idl/start.pro"

    @classmethod
    def set_env(cls, machine):
        '''
        Sets some path for idl/nburst. Implemented only for barth usage
        :param machine: hostname of the machine
        '''
        if machine == 'tycho':
            cls.nburst_working_dir = Path('/obs/blaunet/nburst/')
            cls.idl_binary_path = "/usr/local/idl/bin:/usr/local/idl/bin/bin.linux.x86_64:"
        elif machine == 'barth':
            cls.nburst_working_dir = Path('/Volumes/TOSHIBA/M31/nburst/')
            cls.idl_binary_path = "/usr/local/idl/idl/bin:/usr/local/idl/idl/bin/bin.darwin.x86_64:"
        else:
            pass

    @classmethod
    def from_sitelle_data(cls, axis, spectra, error, fwhm, ORCS_cube, filedir, prefix, force = False):
        """
        This function converts sitelle data to be fitted by Nburst.
        It interpolates a spectra onto a regular wavelength grid, and creates error and fwhm spectra as well.
        Spectra, error and fwhm should have the same **spatial** dimensions (x or x,y)
        :param axis: spectral axis of the cube, in wavenumber [cm-1]
        :param spectra: a 1,2 or 3D spectra dim=[x,y,z]
        :param error: an error array of the same spatial dim as the spectra ([x,y] if spectra is [x,y,z]), i.e one error per pixel
        :param fwhm: a fwhm array of the same spatial dim as the spectra ([x,y] if spectra is [x,y,z]), i.e one fwhm per pixel
        :param ORCS_cube: the SpectralCube from which the spectra is exatracted
        :param filedir: the folder where the NburstFitter should be stored
        :param prefix: a prefix for this data
        """
        if len(spectra.shape) == 1:
            original_spectra = spectra.reshape(1, spectra.shape[0])
        else:
            original_spectra = spectra

        if type(error) is not np.ndarray:
            original_error = np.array([error])
        else:
            original_error = error

        if type(fwhm) is not np.ndarray:
            original_fwhm = np.array([fwhm])
        else:
            original_fwhm = fwhm

        if original_spectra.shape[:-1] != original_error.shape:
            raise ValueError("spatial dimensions don't match : error : %s, spectra : %s"%(original_error.shape, original_spectra.shape))
        if original_spectra.shape[:-1] != original_fwhm.shape:
            raise ValueError("spatial dimensions don't match : fwhm : %s, spectra : %s"%(original_fwhm.shape, original_spectra.shape))

        xlims = ORCS_cube.get_filter_range()
        wl_axis = regular_wl_axis(axis, xlims).astype(float)

        #creation of an error cube
        err_cube = np.repeat(original_error[..., np.newaxis],len(wl_axis), -1)

        #creation of a fwhm cube
        fwhm_cube = np.repeat(original_fwhm[..., np.newaxis],len(wl_axis), -1)
        for xy in np.ndindex(fwhm_cube.shape[:-1]):
            fwhm_cube[xy] = np.array([fwhm*lam**2/1e8 for fwhm,lam in zip(fwhm_cube[xy], wl_axis)])

        #creation of the data cube interpolated on wl axis
        wl_spectra = np.zeros(original_spectra.shape[:-1]+wl_axis.shape)
        for xy in np.ndindex(original_spectra.shape[:-1]):
            wl_spectra[xy] = UnivariateSpline(axis.astype(float),original_spectra[xy],s=0)(1e8/wl_axis)

        #Header
        wl_header = gen_wavelength_header(ORCS_cube.get_header(), wl_axis, len(original_spectra.shape))
        nburst_fitter =  cls(wl_axis, wl_spectra, err_cube, fwhm_cube, wl_header, filedir, prefix, force=force)
        nburst_fitter.fit_params['lmin'] = 4825.0 if ORCS_cube.params.filter_name == 'SN2' else 6480.
        nburst_fitter.fit_params['lmax'] = 5132.0 if ORCS_cube.params.filter_name == 'SN2' else 6842.
        return nburst_fitter

    @classmethod
    def from_sitelle_region(cls, region, ORCS_cube, filedir, prefix, force = False):
        """
        Generate the right parameters to generate a Nburst fittable cube from a region in a SITELLE datacube
        :param region: the region to study
        :param ORCS_cube: SpectralSuce instance
        :param filedir: the file directory where to store the cube
        :param prefix: prefix of this cube
        :param force: (Default False) : force top overwrite an already existing cube with the same name
        """
        axis, spec = ORCS_cube.extract_integrated_spectrum(region)
        error = estimate_noise(axis, spec, ORCS_cube.get_filter_range())
        fwhm = ORCS_cube.get_fwhm_map()[region].mean()
        return cls.from_sitelle_data(axis, spec, error, fwhm, ORCS_cube, filedir, prefix, force)
    @classmethod
    def from_previous(cls, filedir, prefix, fit_name = None):
        """
        Reload a NburstFitter that had been created previously, based on its filedir and prefix.
        :param filedir: the folder where the data is stored
        :param prefix: the prefix of this data
        :param fit_name: (optional) the fit_name, if already performed
        """
        spectra, header = io.read_fits(Path(filedir) / prefix+'_data.fits', return_header = True)
        spectra = spectra.T
        if len(spectra.shape) == 1:
            spectra = spectra.reshape(1, spectra.shape[0])
        errors = io.read_fits(Path(filedir) / prefix+'_error.fits').T
        if len(errors.shape) == 1:
            errors = errors.reshape(1, errors.shape[0])
        fwhms = io.read_fits(Path(filedir)  / prefix+'_fwhm.fits').T
        if len(fwhms.shape) == 1:
            fwhms = fwhms.reshape(1, fwhms.shape[0])
        axis = read_wavelength_axis(header, 1)
        header = swap_header_axis(header, 1, len(spectra.shape))

        return cls(axis, spectra, errors, fwhms, header, filedir, prefix, fit_name)

    # @classmethod
    # def from_single_spectra(cls, axis, spectra, error, fwhm, header, filedir, prefix):
    #     spectra = spectra.reshape(1, spectra.shape[0])
    #     error = error.reshape(1, error.shape[0])
    #     fwhm = fwhm.reshape(1, fwhm.shape[0])
    #     return cls(axis, spectra, error, fwhm, header, filedir, prefix)

    def __init__(self, axis, spectra, error, fwhm, header, filedir, prefix, fit_name=None, force = False):
        """
        Initialize a NBurstFitter instance
        :param axis: a regular wavelength axis in Angstroms
        :param spectra: a 1,2 or 3D spectra of shape [z], [y,z] or [x,y,z]
        :param error: an error array of the same shape as spectra
        :param fwhm: an fwhm array of the same shape as spectra
        :param header: a header describing the data. Should match spectra shape : [NAXIS1, NAXIS2, NAXIS3] <=> [x,y,z]
        :param filedir: the folder where the data should be stored
        :param prefix: the prefix for this data
        :param fit_name: (optional)
        """
        self.axis = axis
        self.spectra = np.atleast_2d(spectra)
        error = np.atleast_2d(error)
        if error.shape != self.spectra.shape:
            raise ValueError("Error shape %s doesn't match spectra shape %s : "%(error.shape, self.spectra.shape))
        else:
            self.error = error
        fwhm = np.atleast_2d(fwhm)
        if fwhm.shape != self.spectra.shape:
            raise ValueError("Fwhm shape %s doesn't match spectra shape %s : "%(fwhm.shape, self.spectra.shape))
        else:
            self.fwhm = fwhm
        self.header = header

        self.fit_params={}
        self.filedir = Path(filedir).abspath()
        self.prefix = prefix
        self.idl_result = None
        self.fitted_spectra = None
        self.fit_name = fit_name

        if not (self.filedir/self.prefix+'_data.fits').isfile() or force is True:
            self._save_input()
        ## Environment variables
        self.template_path = self.nburst_working_dir / 'template.pro'

    def _save_input(self, filedir=None, prefix=None):
        if filedir is None:
            if self.filedir is None:
                raise ValueError('You have to provide a valid directory for the data')
            else:
                filedir = self.filedir
        if prefix is None:
            if self.prefix is None:
                raise ValueError('You have to provide a valid prefix for the data')
            else:
                prefix = self.prefix

        self.filedir = Path(filedir).abspath()
        self.prefix = prefix

        self.filedir.makedirs_p()

        swaped_header = swap_header_axis(copy.deepcopy(self.header), 1, len(self.spectra.shape))

        io.write_fits(self.filedir / self.prefix+'_data.fits', self.spectra.T, swaped_header, overwrite=True)
        io.write_fits(self.filedir / self.prefix+'_fwhm.fits', self.fwhm.T, swaped_header, overwrite=True)
        io.write_fits(self.filedir / self.prefix+'_error.fits', self.error.T, swaped_header, overwrite=True)

    def configure_fit(self, fit_name=None, **kwargs):
        if fit_name is None:
            if self.fit_name is None:
                raise ValueError('Please provide a name for this fit configuration')
            else:
                fit_name = self.fit_name
        self.fit_name = fit_name

        self.fit_params['galfile'] = str(self.filedir / self.prefix+ '_data.fits')
        self.fit_params['errfile'] = str(self.filedir / self.prefix+ '_error.fits')
        self.fit_params['fwhmfile'] = str(self.filedir / self.prefix+ '_fwhm.fits')
        self.fit_params['result_file'] = str(self.filedir / self.prefix+'_'+self.fit_name+'_fitted.fits')
        try:
            nx, ny, nz = self.spectra.shape
            self.fit_params['nx'] = nx
            self.fit_params['nz'] = nz
            self.fit_params['ny'] = ny
            self.fit_params['read3d'] = True
            self.fit_params['xsize'] = ny
        except ValueError:
            nx, nz = self.spectra.shape
            self.fit_params['nx'] = nx
            self.fit_params['nz'] = nz
            self.fit_params['ny'] = 0
            self.fit_params['read3d'] = False
            self.fit_params['xsize'] = False
        if 'silent' not in kwargs:
            kwargs['silent']=True
        self.fit_params.update(kwargs)

        if 'lmin' not in self.fit_params:
            self.fit_params['lmin'] = False
        if 'lmax' not in self.fit_params:
            self.fit_params['lmax'] = False

        if ('vorbin' in self.fit_params and self.fit_params['vorbin'] is True) or ('SN' in self.fit_params and self.fit_params['SN'] is True):
            if ('vorbin' in self.fit_params and 'SN' not in self.fit_params) or ('SN' in self.fit_params and 'vorbin' not in self.fit_params):
                raise ValueError('Please provide a signal to noise ratio value (SN=) for voronoi binning')
            if 'vorbin' in self.fit_params and self.fit_params['read3d'] is False:
                raise ValueError('Voronoi binning can only be performed on a 3d cube')
        if 'vorbin' not in self.fit_params:
            self.fit_params['vorbin'] = False
            self.fit_params['SN'] = False

        if 'exclreg' in self.fit_params:
            if self.fit_params['exclreg'] == 'default':
                self.fit_params.pop('exclreg')

        if 'silent' in self.fit_params and self.fit_params['silent'] is True:
            self.fit_params.pop('silent')
            self.fit_params['plot'] = 0
            self.fit_params['window'] = False


        fit_params = copy.deepcopy(self.fit_params)

        def _assign(key, value):
            if type(value) == bool and value == False:
                return '\n'
            if type(value) == bool and value == True:
                value = 1
            if type(value) == str:
                value = "'"+value+"'"
            return '%s = %s\n'%(key,value)

        with open(self.template_path, 'r') as temp:
            template = temp.readlines()
        for i,line in enumerate(template):
            for k,v in fit_params.items():
                if line.startswith(k):
                    template[i] = _assign(k,v)
                    j = i
                    while '$' in line:
                        j += 1
                        line = template[j]
                        template[j] = _assign(k,False)

                    fit_params.pop(k)

        with open(self.filedir / self.prefix+'_'+self.fit_name+'.pro', 'w') as out:
            out.writelines(template)

    def run_fit(self, silent = False):
        env = os.environ
        env['PATH'] = self.idl_binary_path + env['PATH']
        env['IDL_STARTUP'] = self.idl_startup_script

        script_path = self.filedir / self.prefix+'_'+self.fit_name+'.pro'
        p = subprocess.Popen('idl %s'%script_path, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env)
        if not silent:
            for line in p.stdout.readlines():
                print line,
                sys.stdout.flush()
        retval = p.wait()

        self.read_result()

    def read_result(self, filename=None, force = False):
        if self.fitted_spectra is not None and force is False:
            return None
        if filename is None:
            try:
                filename = self.filedir / self.prefix+'_'+self.fit_name+'_fitted.fits'
            except TypeError:
                filename = self.filedir / self.prefix+'_fitted.fits'
        hdu = fits.open(filename)
        try:
            self.bin_table = hdu[1].data[0][0].astype(np.uint).reshape(self.spectra.shape[:-1])
            self.idl_result = hdu[2].data
        except IndexError:
            print "Can not load results for %s; the fit probably didn't converge"%(filename)
            hdu.close()
            self.idl_result = None
            self.fitted_spectra = None
            return None
        hdu.close()
        specs = self.idl_result['FIT'][self.bin_table]
        idl_axis = self.idl_result['WAVE'][0]
        def interpolate(spec, old_axis, new_axis):
            return UnivariateSpline(old_axis, spec, s=0, ext='zeros')(new_axis)
        self.fitted_spectra = np.apply_along_axis(interpolate, -1, specs, idl_axis, self.axis)
        # coldefs = idl_result.columns
        # coldefs.add_col(fits.Column(name='BESTFIT', array=bestfit.T, format='%dD'%bestfit.shape[0]))
        # self.idl_result = fits.FITS_rec.from_columns(coldefs)

    def extract_spectrum(self, binNumber):
        axis = self.idl_result['WAVE'][binNumber,:]
        fit = self.idl_result['FIT'][binNumber,:]
        return axis,fit

class NburstFitterList():

    nburst_working_dir = Path('/Users/blaunet/Documents/M31/nburst/')
    idl_binary_path = "/usr/local/idl/idl/bin:/usr/local/idl/idl/bin/bin.darwin.x86_64:"
    idl_startup_script = "~/.idl/start.pro"

    @classmethod
    def set_env(cls, machine):
        if machine == 'tycho':
            cls.nburst_working_dir = Path('/obs/blaunet/nburst/')
            cls.idl_binary_path = "/usr/local/idl/bin:/usr/local/idl/bin/bin.linux.x86_64:"
        elif machine == 'barth':
            cls.nburst_working_dir = Path('/Volumes/TOSHIBA/M31/nburst/')
            cls.idl_binary_path = "/usr/local/idl/idl/bin:/usr/local/idl/idl/bin/bin.darwin.x86_64:"
        else:
            pass

    def __init__(self, fitters):
        self.fitters = fitters
        self.axis = self.fitters[0].axis
        self.spectra = np.concatenate([fitter.spectra for fitter in self.fitters], axis=0)
        self.error = np.concatenate([fitter.error for fitter in self.fitters], axis=0)
        self.fwhm = np.concatenate([fitter.fwhm for fitter in self.fitters], axis=0)

        self.fitted_spectra = None
    def configure_fit(self, **kwargs):
        for fitter in self.fitters:
            fitter.configure_fit(**kwargs)

    def run_each(self, silent = False):
        env = os.environ
        env['PATH'] = self.idl_binary_path + env['PATH']
        env['IDL_STARTUP'] = self.idl_startup_script

        script_path = ' '.join([fitter.filedir / fitter.prefix+'_'+fitter.fit_name+'.pro' for fitter in self.fitters])
        p = subprocess.Popen('idl %s'%script_path, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env)
        if not silent:
            for line in p.stdout.readlines():
                print line,
                sys.stdout.flush()
        retval = p.wait()

        self.read_result()
    def read_result(self, force = False):
        if self.fitted_spectra is not None and force is False:
            return None
        for fitter in self.fitters:
            fitter.read_result(force=force)
            if fitter.fitted_spectra is None:
                fitter.fitted_spectra = np.full_like(fitter.spectra, np.nan)
                fitter.idl_result = {}
                for col in self.fitters[0].idl_result.columns:
                    fitter.idl_result[col.name] = np.full_like(self.fitters[0].idl_result[col.name], np.nan)


        self.idl_result={}
        for col in self.fitters[0].idl_result.columns:
            if col.name == 'CHI2' or 'D' in col.format:
                self.idl_result[col.name] = np.concatenate([fitter.idl_result[col.name] for fitter in self.fitters], axis=0)
        self.fitted_spectra = np.concatenate([fitter.fitted_spectra for fitter in self.fitters], axis=0)


if 'johannes' in socket.gethostname() or 'tycho' in socket.gethostname():
    NburstFitter.set_env('tycho')
    NburstFitterList.set_env('tycho')
if 'MacBookAirdeBarthelemy' in socket.gethostname():
    NburstFitter.set_env('barth')
    NburstFitterList.set_env('barth')



from orb.fit import Lines
import pandas as pd
import numpy as np
from scipy.interpolate import UnivariateSpline
from scipy.optimize import curve_fit
from orcs.utils import fit_lines_in_spectrum
from orb.utils.spectrum import line_shift
import gvar

def sky_model_to_remove(mean_spectrum, axis, sky_axis, sky_model):
    """
    This function shifts the velocity of a skymodel to match at best the skylines in a spectrum.
    It returns the fitted vector, that shoukld be then subtracted from the original spectrum.
    :param mean_spectrum: The spectrum containing skylines.
    If it comes from an integrated region, pay attention to give the **mean** spectrum,
    i.e divided by the number of integrated pixels
    :param axis: Axis of the spectrum
    :param sky_axis: axis on which the skymodel is known
    :param skymodel: Sky spectrum to be fitted
    :return: the shifted skymodel, to be removed
    """
    c = 299792.458
    axis = np.log10(axis) #This way, a velocity shift correspond to a constant
                                      #wavenumber offset over the whole axis
    sky_axis = np.log10(sky_axis)

    sky_interpolator = UnivariateSpline(sky_axis, sky_model, s=0)
    fit_vector = np.zeros_like(mean_spectrum)
    scale = np.nanmax(mean_spectrum) - np.nanmedian(mean_spectrum)

    def model(x, h, v):
        shift = -np.log10(1-v/c) # minus sign because it's in wavenumber
        sky_shifted = sky_interpolator(axis.astype(float)+shift)
        return h + UnivariateSpline(axis, sky_shifted, s=0)(x)
    popt, pcov = curve_fit(model, axis, mean_spectrum, p0=[scale,-75.])
    fit_vector = model(axis, *popt)-popt[0]
    return fit_vector


def parse_line_params(rest_lines, fit, error = True, wavenumber = True):
    """
    Parse the fitted parameters to display them as a nice pandas DataFrame
    :param rest_lines: Name or position of the lines at rest
    :param fit: the 'fit_params' in the output of a fit from orcs
    :param error: if True, the errors are also parsed (Default = True)
    :param wavenumber: if True, output is in wavenumber (else in wavelength). Default = True
    :return lines_df: a pandas Dataframe containing all information about the fitted lines
    """
    #Potential names
    lines_cm1 = list()
    names = False
    for iline in rest_lines:
        if isinstance(iline, str):
            names = True
            iline = Lines().get_line_cm1(iline)
        lines_cm1.append(iline)
    lines_cm1 = np.array(lines_cm1)

    lines_list = []
    for i in range(len(lines_cm1)):
        line = {}
        if names:
            line['name'] = rest_lines[i]
        line['rest_pos'] = lines_cm1[i]

        line['height'], line['amp'], line['pos'], line['fwhm'], line['sigma'] = fit['lines_params'][i]
        line['velocity'] = fit['velocity'][i]
        line['snr'] = fit['snr'][i]
        line['flux'] = fit['flux'][i]
        line['broadening'] = fit['broadening'][i]

        if error:
            line['height_err'], line['amp_err'], line['pos_err'], line['fwhm_err'], line['sigma_err'] = fit['lines_params_err'][i]
            line['velocity_err'] = fit['velocity_err'][i]
            line['broadening_err'] = fit['broadening_err'][i]
        lines_list.append(line)

    lines_df = pd.DataFrame(lines_list)
    if not wavenumber:
        lines_df['pos'] = 1e8/lines_df['pos']
        lines_df['rest_pos'] = 1e8/lines_df['rest_pos']
        lines_df['sigma'] = lines_df['sigma']*lines_df['pos']**2/1e8
        if error:
            lines_df['pos_err'] = lines_df['pos_err']*lines_df['pos']**2/1e8
            lines_df['sigma_err'] = lines_df['sigma_err']*lines_df['pos']**2/1e8
    return lines_df

def chi2_func(axis, spectrum, fit_params, delta=8):
    """
    Computes a chi square only around fitted lines
    :param axis: axis of the spectrum
    :param spectrum: spectrum on whoch the fit has been performed
    :param fit_params: fit parameters containing information about the lines (output from an orcs fit)
    :param delta: delta around line peak that should be used to compute the chi2. Default = 8
    :return chi2: a chi square value, computed only around +- sigma * fwhm each line
    """

    #We keep only indices around the fitted lines
    indices_to_keep = []
    for i,line in enumerate(fit_params['fitted_models']['Cm1LinesModel']):
        #indices_to_keep.append(np.argwhere(np.abs(line)>=line.std()/2)[:,0])
        peak_id = np.argmax(line)
        fwhm_pix = fit_params['lines_params'][i][3]/(axis[1]-axis[0])
        imin = max(0, int(round(peak_id-delta*fwhm_pix)))
        imax = min(len(spectrum)-1, int(round(peak_id+delta*fwhm_pix)))
        indices_to_keep.append(np.arange(imin,imax))
    to_keep = np.unique(np.concatenate(indices_to_keep))
    sigma = np.std(spectrum)
    residual = (spectrum-fit_params['fitted_vector'])[to_keep]
    return np.sum(np.square(residual/sigma))

def remove_OH_line(spectrum, theta, cube, **kwargs):
    """
    Function used to remove the strange Halpha line around -400km/s (probably OH line).
    To be used with care, especially when there is a real Halpha signal there.
    :param spectrum: the spectrum to fit
    :param theta: theta value corresponding to the spectrum
    :param cube: Spectral cube instance from which the spctrum is taken
    :return: the spectrium whithout this line (hopefully)
    """
    lines = ['Halpha']

    if 'fmodel' not in kwargs:
        kwargs['fmodel'] = 'sinc'
    if 'pos_def' not in kwargs:
        kwargs['pos_def'] = '1'
    # if 'sigma_cov' not in kwargs:
    #     kwargs['sigma_cov'] = [gvar.gvar(30, 10)]
    # if 'sigma_def' not in kwargs:
    #     kwargs['sigma_def'] = ['1']

    cube._prepare_input_params(lines, **kwargs)
    inputparams = cube.inputparams.convert()
    params = cube.params.convert()
    V_range  = [gvar.gvar(-400, 2)]
    snr_guess = 1

    fit = fit_gas_lines([spectrum, theta], inputparams, params, lines, V_range, snr_guess, silent=True)
    return spectrum - fit['line_spectra']

def fit_gas_lines(z_dim, inputparams, params, lines, V_range, snr_guess = None, silent=False):
    """
    This function fit lines in a given spectrum.
    :param z_dim: 2D-array. z_dim[0] contains the spectrum to fit, z_dim[1] contains theta, the incident angle at which the spectrum was recorded
    :param inputparams: output from cube._prepare_input_params()
    :param params: cube.params.convert()
    :param lines: name of the lines to fit  (understanble by ORCS)
    :param V_range: the range of velocity to test
    :param snr_guess: snr_guess (Default = None)
    :return out: a cube containing at each cell a dict with 'line_spectra': the fitted line spectra,
    'fit_params': the fitetd parameters of the lines,
    'chi2': chi2 list, each elemen,t corresponds to a velocity in V_range
    """
    spectrum = z_dim[0]
    theta = z_dim[1]
    chi2_list = []
    spectra_list = []
    params_list = []
    line_spectra, fit_params, chi2 = None, None, None
    for v in V_range:
        fit = fit_lines_in_spectrum(params, inputparams, 1e10, spectrum,
                              theta, pos_cov=v, snr_guess=snr_guess)
        if fit != []:
            hChi2 = chi2_func(params['base_axis'], spectrum, fit)
            chi2_list.append(hChi2)
            spectra_list.append(np.sum(fit['fitted_models']['Cm1LinesModel'], 0))
            params_list.append(parse_line_params(lines,fit))
            if not silent:
                print 'Init V = ', v, 'Fitted V = ', fit['velocity'][0], 'Chi2 = ', hChi2
    if chi2_list != list():
        imin = np.argmin(chi2_list)
        chi2 = chi2_list[imin]#{'chi2': chi2_list, 'v': V_range[imin]}
        line_spectra = spectra_list[imin]
        fit_params = params_list[imin]
    return {'line_spectra':line_spectra, 'fit_params':fit_params, 'chi2':chi2}



import numpy as np
import orb
from scipy.interpolate import UnivariateSpline


def circular_region(x, y, r):
    """
    Computes a circular region of radius r centered on (x,y)
    :param x: abscissse of the center
    :param y: ordonate of the center
    :param r: radius of the region
    :return region:
    """
    if r < 0: r = 0.001
    X, Y = np.mgrid[0:2048, 0:2064]
    R = np.sqrt(((X-x)**2 + (Y-y)**2))
    return np.nonzero(R <= r)
def square_region(x, y, b):
    """
    Computes a square region of size b*b
    :param x: abscisse of the bottom left corner of the square
    :param y: ordonate of the bottom left corner of the square
    :param b: size of the square
    :return region:
    """
    mask = np.zeros((2048, 2064))
    mask[x:x+b, y:y+b] = 1
    return np.where(mask == 1)
def centered_square_region(x,y,b):
    """
    Computes a square region of size b*b centered on x,y
    :param x: abscisse of the center of the square
    :param y: ordonate of the center of the square
    :param b: size of the square
    :return region:
    """
    mask = np.zeros((2048, 2064))
    if b%2 == 0:
        xmin, xmax = x-b/2, x+b/2
        ymin, ymax = y-b/2, y+b/2
    else:
        xmin, xmax = x-b/2, x+b/2+1
        ymin, ymax = y-b/2, y+b/2+1
    if xmin < 0:
        xmin = 0
    if ymin < 0:
        ymin = 0
    mask[xmin:xmax, ymin:ymax] = 1
    return np.where(mask == 1)

def physical_region(cube, ra, dec, r = 2, circle=True, sex = True, centered=True):
    """
    Computes a physical region on a sitelle cube.
    If circle = True (Default), it's a circulare region centered on ra,dec. Otherwise,
    (ra,dec) are the coordinates of the bottom left corner of a square region
    :param cube: the HDF5 cube
    :param ra: right ascension in sexagesimal if sex = True, else in degrees
    :param dec: declinaison in sexagesimal if sex = True, else in degrees
    :param r: radius of the circle if circle=True, else size of the square
    :param circle: If true, circular region, otherwise square
    :param sex: If true, ra,dec expected to be in sexagesimal, otherwise in degrees
    """
    if sex:
        ra_sex = map(float, ra.split(':'))
        dec_sex = map(float, dec.split(':'))
        ra_deg = orb.utils.astrometry.ra2deg(ra_sex)
        dec_deg = orb.utils.astrometry.dec2deg(dec_sex)
    else:
        ra_deg = ra
        dec_deg = dec
    x, y = map(int, map(round, cube.world2pix((ra_deg, dec_deg))[0]))
    if circle:
        return circular_region(x, y, r)
    elif centered:
        return centered_square_region(x, y, int(round(r)))
    else:
        return square_region(x, y, int(round(r)))

def remap(map, binMap=None, binsize=None, original_shape = None):
    """
    Used to remap a binned map on a full map
    User has to provide either a binMap or a binsize and an original shape
    :param binMap: a map of original size filled with indices corresponding to the bin number in the binned map
    """
    if binMap is None and (binsize is None or original_shape is None):
        raise ValueError("If no binMap is provided, both binsize and original_shape should be")
    if binMap is not None:
        raise  NotImplementedError('binMap not implemented yet')
    else:
        if len(map.shape) > 2:
            original_shape = original_shape + map.shape[2:]
        full_map = np.zeros(original_shape)
        for x,y in np.ndindex(map.shape[:2]):
            full_map[x*binsize:(x+1)*binsize, y*binsize:(y+1)*binsize, ...] = map[x, y, ...]

    return full_map


def get_contour_boundaries(region):
    '''
    Returns the contours of a region
    :param region: tuple of array of indices; result from np.argwhere for example
    :return contour: dict of 4 1-D regions defining the contour : top, bottom, left, right
    '''
    X, Y = region[0], region[1]
    ymin = []
    ymax = []
    for x in np.unique(X):
        x_indices = np.argwhere(X == x)
        ymin.append(Y[x_indices[0]][0])
        ymax.append(Y[x_indices[-1]][0])
    ymin = np.array(ymin)
    ymax = np.array(ymax)

    xmin = []
    xmax = []
    for y in np.unique(Y):
        y_indices = np.argwhere(Y == y)
        xmin.append(X[y_indices[0]][0])
        xmax.append(X[y_indices[-1]][0])
    xmin = np.array(xmin)
    xmax = np.array(xmax)
    contour = {'left': tuple([xmin, np.unique(Y)]),
            'right': tuple([xmax, np.unique(Y)]),
            'top' : tuple([np.unique(X),ymax]),
            'bottom': tuple([np.unique(X),ymin])}

    return contour

def fill_contour(mask, contour):
    """
    Fill the inside of a contour
    :param mask: full map on wich the contour is defined
    :param contour: dict of 4 1D regions : top, bottom, left, right
    :return mask: the mask filled inside the contour
    """
    top_x = contour['top'][0]
    bottom_x = contour['top'][0]
    right_y = contour['right'][1]
    left_y = contour['left'][1]
    X_range = top_x if len(top_x) >= len(bottom_x) else bottom_x
    Y_range = right_y if len(right_y) >= len(left_y) else left_y
    for i,x in enumerate(X_range):
        for j, y in enumerate(Y_range):
            if x >= contour['left'][0][j] and x <= contour['right'][0][j]:
                if y >= contour['bottom'][1][i] and y <= contour['top'][1][i]:
                    mask[x,y] = 1
    return mask

def smooth_contour(contour):
    """
    Smooth discontinuous contours
    :param contour: dict of 4 1D regions : top, bottom, left, right
    :return contour: continuous contour
    """
    new_contour = {}
    _, a_left = contour['left']
    _, a_right = contour['right']
    a_h = np.arange(min(a_left.min(), a_right.min()), max(a_left.max(), a_right.max())+1)
    a_top, _ = contour['top']
    a_bottom, _ = contour['bottom']
    a_v = np.arange(min(a_top.min(), a_bottom.min()), max(a_top.max(), a_bottom.max())+1)
    for k in contour:
        if k in ['left', 'right']:
            O, A = contour[k]
        else:
            A, O = contour[k]
        #Make sure that abscisse is unique and sorted
        A, unique_id = np.unique(A, return_index=True)
        O = O[unique_id]
        argsort_A = np.argsort(A)
        smoother = UnivariateSpline(A[argsort_A],O[argsort_A], s=0)
        if k in ['left', 'right']:
            new_A = a_h
        else:
            new_A = a_v
        #new_A = np.arange(A.min(), A.max()+1)
        new_O = np.apply_along_axis(lambda x: map(int, map(round, x)), 0, smoother(new_A))

        if k in ['left', 'right']:
            new_contour[k] = tuple([new_O, new_A])
        else:
            new_contour[k] = tuple([new_A, new_O])

    return new_contour

def convert_contour(original_cube,  contour, new_cube):
    """
    Convert a contour computed on a given cube to a new cube
    :param original_cube: the one on which the contour has been computed
    :param new_cube: the one on which we want to project it
    :param contour: the contour, i.e. a dict of 4 1D  regions : top, bottom, left, right
    :return contour: the contour projected on the new cube
    """
    new_contour = {}
    for k in contour.keys():
        newx = []
        newy = []
        for xy in zip(contour[k][0], contour[k][1]):
            xi,yi =  new_cube.world2pix(original_cube.pix2world(xy))[0]
            if xi < 0:
                xi = 0
            elif xi > new_cube.dimx-1:
                xi = new_cube.dimx-1
            if yi <0:
                yi = 0
            elif yi > new_cube.dimy-1:
                yi = new_cube.dimy-1
            newx.append(int(round(xi)))
            newy.append(int(round(yi)))
        new_contour[k] = tuple([np.array(newx), np.array(newy)])
    return new_contour



import matplotlib.pyplot as plt
import numpy as np
import matplotlib.patches as patches
from orb.fit import Lines
from photutils import CircularAperture

def customize_axes(axes, **kwargs):
    for k,v in kwargs.items():
        try:
            getattr(axes, 'set_%s'%k)(v)
        except AttributeError:
            raise AttributeError(str(type(axes))+' has no method set_%s'%k)
def make_wavenumber_axes(ax, **kwargs):
    customize_axes(ax, **kwargs)
    _make_spectral_axes(ax, False)
def make_wavelength_axes(ax, **kwargs):
    customize_axes(ax, **kwargs)
    _make_spectral_axes(ax, True)

def _make_spectral_axes(ax, wavelength):
    wl_label = 'Wavelength [Angstroms]'
    wn_label = "Wavenumber [cm$^{-1}$]"
    if wavelength:
        bottom_label = wl_label
        top_label = wn_label
    else:
        bottom_label = wn_label
        top_label = wl_label
    customize_axes(ax, xlabel= bottom_label,
                       ylabel='Flux [erg/cm$^2$/s/A]')
    #Angstroms x axis
    ax2=ax.twiny()
    customize_axes(ax2, xlim=ax.get_xlim(),
                        xticks = ax.get_xticks()[1:-1],
                        xticklabels = ["%.f" % (1e8/wn) for wn in ax.get_xticks()[1:-1]],
                        xlabel = top_label)

def lines_pos(lines_name, v, wavenumber=False):
    if wavenumber is True:
        return [(wn*(1+v/3e5)) for wn in Lines().get_line_cm1(lines_name)]

    else:
        return [1e8/(wn*(1-v/3e5)) for wn in Lines().get_line_cm1(lines_name)]

def add_lines_label(ax, filter, velocity, wavenumber=False,offset=15):
    if type(velocity) != tuple and type(velocity) != list:
        velocity = [velocity]
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    if filter == 'SN2':
        lines_names = ['Hbeta','[OIII]4959', '[OIII]5007']
    elif filter == 'SN3':
        lines_names = ['[NII]6548', 'Halpha', '[NII]6583', '[SII]6716', '[SII]6731']
    pos = lines_pos(lines_names, velocity[0], wavenumber)

    for i, name in enumerate(lines_names) :
        #ax.annotate(name, ((pos[i]-offset-xmin)/(xmax-xmin), 0.99), xycoords='axes fraction', rotation=90.)
        ax.annotate(name, (pos[i]-offset, 0.99*ymax), rotation=90.,  annotation_clip = False)
        #ax.text((pos-10-xmin)/(xmax-xmin), 0.94, name, rotation = 45.)

        color = iter(['k', 'r', 'g'])
        for v in velocity:
            ax.axvline(lines_pos(lines_names, v, wavenumber)[i], ymin=0.95, c=next(color), ls='-', lw=1.)

## Plot a 2D map
def plot_map(data, ax=None, region=None, projection=None,
                colorbar=False,
                xlims=None, ylims=None,
                **kwargs):
    """
    Helper function to plot 2D maps

    :param map: the 2d map to plot
    :param region: a region (obtained with np.where() for example) to plot on top of the image
    :param projection: a WCS projection to plot the map on
    :param colorbar: if True (non default), a color bar is associated to the plot
    :param pmin: (Optional) if passed, vmin set to np.nanpercentile(data, pmin)
    :param pmax: (Optional) if passed, vmax set to np.nanpercentile(data, pmax)
    :param kwargs: kwargs passed to imshow() function (e.g vmin, cmap etc..)
    """
    if ax is None: #No ax has been given : we have to create a new one
        if projection:
            fig, ax = plt.subplots(subplot_kw={'projection':projection})
        else:
            fig,ax = plt.subplots()
    else:
        fig = ax.get_figure()
        if projection: #We have to remove the current ax and replace it with one handling the projection
            pos = ax.get_position()
            ax.remove()
            ax = fig.add_axes(pos, projection=projection)

    if projection:
        ax.coords[0].set_major_formatter('hh:mm:ss.s')
        ax.coords[1].set_major_formatter('dd:mm:ss')
        ax.set_xlabel('RA')
        ax.set_ylabel('DEC')
    else:
        ax.set_xlabel('X')
        ax.set_ylabel('Y')

    pmin = kwargs.pop('pmin', None)
    pmax = kwargs.pop('pmax', None)
    if pmin is not None or pmax is not None:
        if pmin is None:
            pmin = 10
        if pmax is None:
            pmax = 90
        kwargs['vmin'] = np.nanpercentile(data, pmin)
        kwargs['vmax'] = np.nanpercentile(data, pmax)

    cmap = kwargs.pop('cmap', 'gray_r')
    origin = kwargs.pop('origin', 'lower')
    im = ax.imshow(data.T, origin=origin, cmap=cmap, **kwargs)

    if region:
        if isinstance(region[0], tuple): #We have a list of regions:
            for r in region:
                mask = np.zeros_like(data)
                mask[r] = True
                ax.contour(mask.T, 1, colors='r')
        else:
            mask = np.zeros_like(data)
            mask[region] = True
            ax.contour(mask.T, 1, colors='r')
    if colorbar:
        fig.colorbar(im)
    if projection:
        ax.grid()
    return fig, ax

def plot_scatter(x,y,ax=None, **kwargs):
    """
    Helper function to plot scattered data
    :param x: x coordinates
    :param y: y coordinates
    :param ax: the matplotlib ax on which to plot. If None, a new one is created
    :param label: (Optional) labels of the data
    :param color: (Default 'red') color of the markers
    :param marker: (Default '+') Marker to use
    :param **kwargs: any keyword argument accepted by plt.scatter()
    :return fig, ax:
    """
    if ax is None: #No ax has been given : we have to create a new one
        fig,ax = plt.subplots()
    else:
        fig = ax.get_figure()
    label = kwargs.pop('label', None)
    color = kwargs.pop('c', 'red')
    color = kwargs.pop('color', color)
    marker = kwargs.pop('marker', '+')
    ax.scatter(x,y,label = label, color = color, marker = marker, **kwargs)
    if label is not None:
        ax.legend()
    return fig, ax
def plot_spectra(axis, spectrum, ax=None, wavenumber=True, **kwargs):
    """
    Helper function to plot a spectrum
    :param axis: the spectrum axis
    :param spectrum: the spectrum itself
    :param label: (Optional) label of the plot
    :param wavenumber: (Default True) if the axis is in wavenumber ([cm-1]) or wavelength ([Angstroms])
    :param ax: ax where to plot. if none, a new one is created
    :param kwargs: Any kwargs accepted by plt.plot()

    """
    if ax is None: #No ax has been given : we have to create a new one
        fig,ax = plt.subplots()
    else:
        fig = ax.get_figure()
    label = kwargs.pop('label', None)
    ax.plot(axis, spectrum, label = label, **kwargs)
    if label is not None:
        ax.legend()

    if wavenumber:
        make_wavenumber_axes(ax)
    else:
        make_wavelength_axes(ax)
    return fig, ax

def plot_hist(map, ax=None, log = False, pmin = None, pmax=None, **kwargs):
    """
    Helper function to plot an histogram.
    Especially helpful when dealing with 2d values (map) containing NaN (they are excluded from the analysis)
    :param map: the data from which the histogram is taken
    :param ax: ax where to plot. If None, a new one is created
    :param pmin: (Optional) if passed, data is cut to np.nanpercentile(map, pmin)
    :param pmax: (Optional) if passed, vmin set to np.nanpercentile(map, pmax)
    :param kwargs: Any kwargs accepted by np.histogram
    """
    _map = np.copy(map)
    if type(map) is np.ma.MaskedArray:
        _map = _map[~map.mask]
    if ax is None:
        f,ax = plt.subplots()
    else:
        f = ax.get_figure()
    if pmin is not None:
        min = np.nanpercentile(_map, pmin)
    else:
        min = np.nanmin(_map)
    if pmax is not None:
        max = np.nanpercentile(_map, pmax)
    else:
        max = np.nanmax(_map)
    _map[_map > max] = np.nan
    _map[_map < min] = np.nan
    h = np.histogram(_map[~np.isnan(_map)], **kwargs)
    X = h[1][:-1]
    Y = h[0]
    ax.bar(h[1][:-1], h[0], align='edge', width = h[1][1]-h[1][0], log=log)
    ax.set_title('Median : %.2e, Std : %.2e'%(np.nanmedian(_map), np.nanstd(_map)))
    return f,ax

def plot_sources(sources, ax, **kwargs):
    """
    Helper function to plot sources nicely over a map
    WARNING : sources should respect astropy convention : x and y reversed
    :param sources: a Pandas Dataframe containing at least columns 'xcentroid' and 'ycentroid'
    :param ax: the ax on which to plot (should be containing the map on top of which we are going to plot the sources)
    :param kwargs: any kwargs accepted by CircularAperture.plot
    """
    f = ax.get_figure()
    positions=(sources['ycentroid'], sources['xcentroid'])
    apertures = CircularAperture(positions, r=4.)
    color = kwargs.pop('c', 'red')
    color = kwargs.pop('color', color)
    lw = kwargs.pop('lw', 1.5)
    alpha = kwargs.pop('alpha', 0.5)
    apertures.plot(color=color, lw=lw, alpha=alpha, ax=ax, **kwargs)
    return f,ax

class Interactive1DPlotter:
    def __init__(self, axes, cube_axis, cube, *args, **kwargs):
        self.axes = axes
        self.cube_axis = cube_axis
        self.cube = cube
        self.args = args
        self.kwargs = kwargs

        self.patch = None
        self.annotation = None
        self.artist = []
    def connect(self, figure):
        'connect to all the events we need'
        self.figure = figure
        self.cidpress = self.figure.canvas.mpl_connect(
            'button_press_event', self.on_press)
        self.cidmotion = self.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)

    def on_press(self, event):
        'on button press we will see if the mouse is over us and store some data'
        if event.inaxes is None: return
        x, y = map(int, map(round, (event.xdata, event.ydata)))
        if self.artist is not []:
            for a in self.artist:
                a.remove()
        self.axes.get_figure().show()
        self.artist = self.axes.plot(self.cube_axis, self.cube[x,y,...], *self.args, **self.kwargs)
        # for ax in self.figure.get_axes():
        #     ax.relim()
        #     ax.autoscale_view()
        self.axes.relim()
        self.axes.autoscale_view()
        self.axes.legend()
        self.axes.get_figure().show()

    def on_motion(self, event):
        if event.inaxes is None: return

        x, y = map(int,map(round, (event.xdata, event.ydata)))

        if self.patch is None:
            ax = self.figure.axes[0]
            self.patch = patches.Rectangle((x-0.5,y-0.5),1,1,fill=False, ec='w')
            ax.add_patch(self.patch)
            data = ax.images[0].get_array()
            self.annotation = ax.annotate('V = %f km/s'%(data.T[x,y]) ,(0,1.1), xycoords='axes fraction' )
        elif self.patch.xy == (x-0.5,y-0.5): return
        else:
            self.patch.remove()
            self.patch = None
            self.annotation.remove()


        self.figure.canvas.draw()


class SpectraPlotter:
    def __init__(self, axis, original_cube, fit_cube, plot_axis, projection=None, residual = True, **kwargs):
        self.axis = axis
        self.original_cube = original_cube
        self.fit_cube = fit_cube
        self.residual = residual
        self.plot_axis = plot_axis
        self.projection = projection
        self.kwargs = kwargs

        self.patch = None
        self.annotation = None
        self.annotation2 = None
    def connect(self, figure):
        'connect to all the events we need'
        self.figure = figure
        self.cidpress = self.figure.canvas.mpl_connect(
            'button_press_event', self.on_press)
        self.cidmotion = self.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)

    def on_press(self, event):
        'on button press we will see if the mouse is over us and store some data'
        if event.inaxes is None: return
        x, y = map(int, map(round, (event.xdata, event.ydata)))
        self.plot_axis.clear()
        self.plot_axis.get_figure().show()

        args = (self.axis, self.original_cube[x,y,:],self.axis, self.fit_cube[x,y,:])
        if self.residual:
            args += (self.axis, self.original_cube[x,y,:]-self.fit_cube[x,y,:])
        plot_spectra(*args, ax = self.plot_axis, **self.kwargs)
        self.plot_axis.get_figure().show()

    def on_motion(self, event):
        if event.inaxes is None: return

        x, y = map(int,map(round, (event.xdata, event.ydata)))

        if self.patch is None:
            ax = self.figure.axes[0]
            self.patch = patches.Rectangle((x-0.5,y-0.5),1,1,fill=False, ec='w')
            ax.add_patch(self.patch)
            data = ax.images[0].get_array()
            self.annotation = ax.annotate('V = %f km/s'%(data.T[x,y]) ,(0,1.1), xycoords='axes fraction' )
            if self.projection is not None:
                ra, dec = self.projection.pix2world((x*48+24, y*48+24))[0]
                self.annotation2 = ax.annotate('RA = %s \nDEC = %s'%(ra,dec) ,(0,1.15), xycoords='axes fraction' )
        elif self.patch.xy == (x-0.5,y-0.5): return
        else:
            self.patch.remove()
            self.patch = None
            self.annotation.remove()
            if self.projection is not None:
                self.annotation2.remove()




        self.figure.canvas.draw()

    def disconnect(self):
        'disconnect all the stored connection ids'
        self.figure.canvas.mpl_disconnect(self.cidpress)



import numpy as np
from scipy.interpolate import UnivariateSpline
import orb
def estimate_noise(full_axis, full_spectrum, filter_lims, side='both'):
    '''
    Estimates the noise in an integrated spectra from sitelle, by measuring it outside of the filter range.
    :param full_axis: full wavenumber axis of the cube
    :param full_spectrum: spectrum on this full axis
    :param filter_lims: limits of the filter range (in wavenumber)
    :param side: (optional) if 'right' ('left'), only the right (left) side of the filter is considered. Default to 'both'
    :return: the standard deviaition of the noise outside the filter range
    '''
    imin,imax = np.searchsorted(full_axis,filter_lims)
    if side is 'right':
        noise = full_spectrum[imax+40:]
    elif side is 'left':
        noise = full_spectrum[:imin-40]
    elif side is 'both':
        noise = np.concatenate((full_spectrum[:imin-40], full_spectrum[imax+40:]))
    else:
        raise ValueError(side)
    return np.std(noise)



def gen_wavelength_header(h, axis, ax_num = 3):
    """
    Generates a FITS wavelength header in Ansgtroms
    :param h: the header to update
    :param axis: the regular wavelength axis (numpy array or list)
    :param ax_num: (Default 3) The axis number to use in the header
    :return: the updated header
    """
    h['NAXIS%d'%ax_num] = len(axis)
    h['CRPIX%d'%ax_num] = 1
    h['CRVAL%d'%ax_num] = axis[0]
    h['CDELT%d'%ax_num] = axis[1]-axis[0]
    h['CUNIT%d'%ax_num] = 'Angstroms'
    h['CTYPE%d'%ax_num] = 'LINEAR'
    return h

def read_wavelength_axis(header, axis):
    """
    Generates the right wavelength axis from a FITS header
    :param header: the header to consider
    :param axis: the number of the axis on which the wavelength axis is stored in the header
    :return: a numpy array describing the waveklength axis
    """
    offset = 1-header['CRPIX'+str(axis)]
    grid =  np.arange(offset,header['NAXIS'+str(axis)] + offset)*header['CDELT'+str(axis)]
    return header['CRVAL'+str(axis)] + grid

def rebin(map, binsize, type):
    xsize, ysize = map.shape[:2]
    new_xsize = xsize//binsize if xsize % binsize == 0 else xsize//binsize + 1
    new_ysize = ysize//binsize if ysize % binsize == 0 else ysize//binsize + 1
    rebinned = np.zeros((new_xsize, new_ysize)+ map.shape[2:])

    for x,y in np.ndindex((new_xsize, new_ysize)):
        data = map[x*binsize:(x+1)*binsize, y*binsize:(y+1)*binsize, ...]
        if type == 'ERR':
            rebinned[x,y, ...] = np.sqrt( np.nansum(  np.square( data )))
        else:
            rebinned[x, y, ...] = np.nanmean(data, (0,1))
    return rebinned

def regular_wl_axis(axis, xlims=None):
    """
    Converts a wavenumber axis ([cm-1]) into a regular (==equally spaced) wavelength axis ([Angstroms])
    :param axis: the input axis in cm-1
    :param xlims: optional : limits in cm-1 to reduce the range of the axis (typically : filter range)
    :return reg_axis: a regular axis in Angstroms
    """
    irreg_axis = 1e8/axis
    reg_axis = np.linspace(irreg_axis[0], irreg_axis[-1], axis.shape[0] )
    #We flip it to make it wavelength ascending
    reg_axis = np.flip(reg_axis, 0)

    if xlims:
        #We get extremas of the filter bandpass
        wl_min, wl_max = list(reversed([1e8/x for x in xlims]))
        #We conserve only spectrum inside this bandpass
        i_min = np.searchsorted(reg_axis, wl_min)
        i_max = np.searchsorted(reg_axis, wl_max)
        reg_axis = reg_axis[i_min:i_max]
    return reg_axis

def wavelength_regrid(cube, rebinned, only_bandpass, type, header=None):
    ## CREATION OF THE NEW AXIS
    base_axis = cube.params.base_axis.astype(float)
    irreg_axis = 1e8/base_axis
    reg_axis = np.linspace(irreg_axis[0], irreg_axis[-1], base_axis.shape[0] )
    #We flip it to make it wavelength ascending
    reg_axis = np.flip(reg_axis, 0)

    if only_bandpass:
        #We get extremas of the filter bandpass
        wl_min, wl_max = list(reversed([1e8/x for x in cube.params.filter_range]))
        #We conserve only spectrum inside this bandpass
        i_min = np.searchsorted(reg_axis, wl_min)
        i_max = np.searchsorted(reg_axis, wl_max)
        reg_axis = reg_axis[i_min:i_max]
    new_axis = 1e8/reg_axis #back in cm-1 to evaluate the spectrum function on it


    ##INTERPOLATION OF THE CUBE
    wl_cube = np.zeros((rebinned.shape[0], rebinned.shape[1], len(new_axis)))

    def interpolator(spectrum, old_axis, new_axis, type):
        f = UnivariateSpline(old_axis, spectrum, s=0)
        res = f(new_axis)
        if type == 'FWHM':
            res = np.array([r*1e8/n**2 for r,n in zip(res,new_axis)])
        return res

    if type != 'ERR':
        wl_cube = np.apply_along_axis(interpolator, 2, rebinned,
                                                    old_axis=base_axis,
                                                    new_axis=new_axis,
                                                    type = type)
    else:
        wl_cube = rebinned[:,:,:len(new_axis)]

    ## header
    if header is None:
        header = cube.get_header()
    h = gen_wavelength_header(header, reg_axis)

    return wl_cube, h

def nm_to_cm1(spectrum, nm_axis, out_axis):
    cm1_axis = np.flip(1e8/nm_axis.astype(float),0)
    spectrum = np.flip(spectrum, 0)
    cm1_spectrum = UnivariateSpline(cm1_axis, spectrum.astype(float), s=0, k=1,ext=1)(out_axis.astype(float))
    return cm1_spectrum

def swap_header_axis(h, a0, a1):
    keywords = ['NAXIS', 'CRPIX', 'CRVAL', 'CDELT', 'CUNIT', 'CTYPE']
    tmp = {}
    for k in keywords:
        try:
            tmp[k] = h[k+str(a0)]
        except KeyError:
            tmp[k] = ''
        try:
            h[k+str(a0)] = h[k+str(a1)]
        except KeyError:
            h[k+str(a0)] = ''
        h[k+str(a1)] = tmp[k]

    for k in h.keys():
        if 'PC' in k:
            if str(a0) in k:
                h[k.replace(str(a0), str(a1))] = h.pop(k)
    return h

def filter_star_list(_star_list):
    """
    Very basic filter to remove stars position which are out of the image.
    Hard coded to an image of 2048*2064 pix
    :param _star_list: the list of XY pixel coordinates
    :return: the filtered list
    """
    _star_list = np.copy(_star_list)
    for istar in range(_star_list.shape[0]):
        if (_star_list[istar,0] < 0
            or _star_list[istar,0] > 2048
            or _star_list[istar,1] < 0
            or _star_list[istar,1] > 2064):
            _star_list[istar,:] = np.nan
    return _star_list
def measure_dist(pos1,pos2):
    """
    Measure the distance (norm 2) between two positions, in units of the position
    :param pos1: the first position
    :param pos2: the second position
    :return: the distance (norm 2)
    """
    return np.sqrt((pos1[:,0] - pos2[:,0])**2+(pos1[:,1] - pos2[:,1])**2)
def get_star_list(star_list_deg, im, hdr, dxmap, dymap):
    """
    Computes the pixel positions of a list of positions in degrees.
    :param star_list_deg: a 2d array of object positions, in degree
    :param im: the image on which this objects are supposed to be
    :param hdr: the header of this image
    :param dxmap: the dxmap correction to apply
    :param dymap: the dymap correction to apply
    :return: two lists of pixel poition : one that includes dxdymaps in the copmputation, one that doesn't
    
    """
    #Star positions without dxdymaps
    dxmap_null = np.copy(dxmap)
    dxmap_null.fill(0.)
    dymap_null = np.copy(dymap)
    dymap_null.fill(0.)
    star_list_pix1 = orb.utils.astrometry.world2pix(
        hdr, im.shape[0], im.shape[1], np.copy(star_list_deg), dxmap_null, dymap_null)
    star_list_pix1 = filter_star_list(star_list_pix1)

    #Star positions with dxdymaps
    star_list_pix2 = orb.utils.astrometry.world2pix(
        hdr, im.shape[0], im.shape[1], np.copy(star_list_deg), dxmap, dymap)

    star_list_pix2 = filter_star_list(star_list_pix2)

    return star_list_pix1, star_list_pix2



import multiprocessing
import numpy as np
from orb.utils.parallel import init_pp_server, close_pp_server

def parallel_apply_along_axis(func1d, axis, arr, *args, **kwargs):
    """
    Like numpy.apply_along_axis(), but takes advantage of multiple
    cores.
    -------------
    Additional keywords :
    :param modules: modules to be imported so that the function works correctly. Example : 'import numpy as np'
    :param depfuncs: functions used by the main function but defined outside of its body
    """
    job_server, ncpus = init_pp_server(multiprocessing.cpu_count(),silent=False)

    chunks = [(func1d, axis, sub_arr, args, kwargs)
              for sub_arr in np.array_split(arr, ncpus, axis=((axis + 1 )% 3))]

    def helper(func1d, axis, arr, args, kwargs):
        return np.apply_along_axis(func1d, axis, arr, *args, **kwargs)

    modules = kwargs.pop('modules', tuple())
    modules += ('import numpy as np',)

    depfuncs = kwargs.pop('depfuncs', tuple())
    depfuncs += (func1d,)
    jobs = [job_server.submit(
                helper,
                args=(c),
                modules=(modules),
                depfuncs=(depfuncs))
                for c in chunks]

    job_output = [job() for job in jobs]
    close_pp_server(job_server)

    return np.concatenate(job_output)

def parallel_apply_over_frames(func2d, cube, *args, **kwargs):
    """
    Iterates over the z dimension of a cube to apply a function on each frame,
    taking advantage of multiple cores.

    :param func2d: the function to apply to each frame
    :param cube: the data cube
    :param modules: modules to be imported so that the function works correctly. Example : 'import numpy as np'
    :param depfuncs: functions used by the main function but defined outside of its body
    :param args: arguments to be passed to func2d
    :param kwargs: keyword arguments to be passed to func2d
    """
    try:
        if len(cube.shape) != 3:
            raise ValueError('Shape of the cube must be 3 dimensional; got %s instead'%(cube.shape))
    except AttributeError as e:
        raise TypeError('Incorrect type for the data cube : %s'%type(a))


    job_server, ncpus = init_pp_server(multiprocessing.cpu_count(),silent=False)
    if cube.shape[-1]<ncpus:
        ncpus=cube.shape[-1]
    chunks = [(func2d, sub_cube, args, kwargs)
              for sub_cube in np.array_split(cube, ncpus, axis=2)]

    def helper(func2d, cube, args, kwargs):
        #We have to determine dimension of what is returned
        res0 = np.array(func2d(cube[:,:,0], *args, **kwargs))
        return cube
        # return res0.shape
        # res_cube = np.zeros((res0.shape + (cube.shape[-1], )))
        # for i in range(1,(cube.shape[-1])):
        #     res_cube[...,i] = func2d(cube[:,:,i], *args, **kwargs)
        # return res_cube

    modules = kwargs.pop('modules', tuple())
    modules += ('import numpy as np',)

    depfuncs = kwargs.pop('depfuncs', tuple())
    depfuncs += (func2d,)
    jobs = [job_server.submit(
                helper,
                args=(c),
                modules=(modules),
                depfuncs=(depfuncs))
                for c in chunks]

    job_output = [job() for job in jobs]
    close_pp_server(job_server)
    return job_output
    try:
        return np.concatenate(job_output, axis=-1)
    except Exception as e:
        print 'Could not concatenate'
        return job_output




# def parallel_loop(func, iterator, *args, **kwargs):
#     job_server, ncpus = init_pp_server(multiprocessing.cpu_count(),silent=False)
#     it_list = list(iterator)
#     it_chunks = [it_chunk for it_chunk in np.array_split(it_list, ncpus)]
#
#     def helper(func, iterator, args, kwargs):
#         out = []
#         for i in iterator:
#             out.append(func(i, *args, **kwargs))
#         return out
#
#     modules = kwargs.pop('modules', tuple())
#     modules += ('import numpy as np',)
#
#     depfuncs = kwargs.pop('depfuncs', tuple())
#     depfuncs += (func,)
#     jobs = [job_server.submit(
#                 helper,
#                 args=(func, it_chunk, args, kwargs),
#                 modules=(modules),
#                 depfuncs=(depfuncs))
#                 for it_chunk in it_chunks]
#
#     job_output = []
#     for job in jobs:
#         job_output.append(job())
#
#     close_pp_server(job_server)
#     return np.array(job_output).flatten()



from astropy.coordinates import SkyCoord
import astropy.units as u
from path import Path
import socket

M31_CENTER = SkyCoord(ra='00h42m44.371s', dec='41d16m08.34s')#, distance=778*u.kpc)

if 'MacBookAirdeBarthelemy' in socket.gethostname():
    FITS_DIR = Path('/home/blaunet/Documents/M31/fits/')

if 'johannes' in socket.gethostname() or 'tycho' in socket.gethostname():
    FITS_DIR = Path('/data/blaunet/fits/')
if 'lp-alm-stg' in socket.gethostname():
    FITS_DIR = Path('/Users/blaunet/Documents/M31/fits')



from setuptools import setup, find_packages
from sitelle import version

packages = find_packages(where=".")
setup(
    name="sitelle",
    packages=packages,
    url='https://gitlab.ethz.ch/blaunet/sitelle',
    author='Barthelemy Launet',
    author_email='barthelemy.launet@obspm.fr',
    maintainer='Barthelemy Launet',
    maintainer_email='barthelemy.launet@obspm.fr',
    version=version.__version__,
    description='Helper functions to work with M31 observations on SITELLE',
    include_package_data=False,
    platforms='any',
)
